import copy


def deepupdate(target, src):
    """Deep update target dict with src
    For each k,v in src: if k doesn't exist in target, it is deep copied from
    src to target. Otherwise, if v is a list, target[k] is extended with
    src[k]. If v is a set, target[k] is updated with v, If v is a dict,
    recursively deep-update it.

    Examples:
    >>> t = {'name': 'toto', 'hobbies': ['programming', 'chess']}
    >>> deepupdate(t, {'hobbies': ['gaming']})
    >>> print t
    {'name': 'toto', 'hobbies': ['programming', 'chess', 'gaming']}
    """
    from pylogctx.core import LazyAccessor
    if src and isinstance(src, dict):
        for k, v in src.items():
            if k in target and target[k] is not None and isinstance(
                    v, (list, dict, set)):
                if isinstance(v, list):
                    target[k].extend(v)
                elif isinstance(v, dict):
                    deepupdate(target[k], v)
                elif isinstance(v, set):
                    target[k].update(v)
            else:
                if isinstance(v, LazyAccessor):
                    # To have a lazy representation from the instance we need
                    # to keep the reference
                    target[k] = v
                elif isinstance(v, dict):
                    # For manage nested LazyAccessor object
                    target[k] = {}
                    deepupdate(target[k], v)
                else:
                    target[k] = copy.deepcopy(v)



from __future__ import absolute_import

import logging
import warnings

from celery import current_task
from celery import Task

from .compat import getargspec
from .core import context


logger = logging.getLogger(__name__)


class LoggingTask(Task):

    def before_call(self, *args, **kwargs):
        # Override if you need to add some vars or log something.
        pass

    def after_call(self, *args, **kwargs):
        # Override if you need to add some vars or log something.
        pass

    def __call__(self, *args, **kwargs):
        # Save context to put in back at the end. Useful when a task call
        # another task to avoid context from another task.
        self.save_context = context.as_dict()
        context.clear()

        arg_spec = getargspec(self.before_call)
        if arg_spec.varargs is None:
            # To keep breaking compat
            warnings.warn('Method `before_call` without args is deprecated')
            self.before_call()
        else:
            self.before_call(*args, **kwargs)

        task = current_task

        try:
            context.update(task)
        except Exception as e:
            logger.debug('Failed to push task to log context: %r', e)

        try:
            return super(LoggingTask, self).__call__(*args, **kwargs)
        finally:
            arg_spec = getargspec(self.after_call)
            if arg_spec.varargs is None:
                # To keep breaking compat
                warnings.warn('Method `after_call` without args is deprecated')
                self.after_call()
            else:
                self.after_call(*args, **kwargs)
            context.clear()
            if self.save_context:
                context.update(**self.save_context)



import inspect


def getargspec(callable):
    try:
        # Python 3
        return inspect.getfullargspec(callable)
    except AttributeError:
        # Python 2
        return inspect.getargspec(callable)



from __future__ import absolute_import

import logging
from django.conf import settings
from django.utils.deprecation import MiddlewareMixin
from pylogctx import context, AdapterNotFound

logger = logging.getLogger(__name__)


class OuterMiddleware(MiddlewareMixin):
    def process_request(self, request):
        context.clear()
        try:
            context.update(request)
        except AdapterNotFound:
            logger.info("Can't adapt %s for log.", request.__class__.__name__)

    def process_response(self, request, response):
        context.clear()
        return response

    def process_exception(self, request, exception):
        context.clear()
        return None


class ExtractRequestContextMiddleware(OuterMiddleware):
    def process_request(self, request):
        context.clear()
        extractor = settings.PYLOGCTX_REQUEST_EXTRACTOR
        try:
            context.update(**extractor(request))
        except Exception:
            logger.exception()



from .core import (
    AdapterNotFound,
    AddContextFilter,
    AddContextFormatter,
    ExcInfoFilter,
    LazyAccessor,
    context,
    log_adapter,
)

__all__ = [
    'AdapterNotFound',
    'AddContextFilter',
    'AddContextFormatter',
    'ExcInfoFilter',
    'LazyAccessor',
    'context',
    'log_adapter',
]



from __future__ import unicode_literals

import itertools
import logging
import threading

from .helpers import deepupdate


_log = logging.getLogger(__name__)


class Context(threading.local):
    @property
    def data(self):
        # Since we are local, the property is not initialize upon init on
        # object (which is shared) but on first access *in* thread.
        try:
            self._stack
        except AttributeError:
            self._stack = [{}]
        return self._stack[0]

    def clear(self):
        """
        Drop all records
        """
        self.data.clear()

    def update(self, *objects, **fields):
        """
        Push records in context.

        `objects`: Instance of few/an object(s), usually a model object.

        `fields`: Custom fields to push in context.

        Usage:
        .. code-block:: python

            log_context.update(Request, Task)
            log_context.update(Request, userId=2, articleId=4)
        """
        deepupdate(self.data, fields)

        for object_ in objects:
            if not object_:
                # You can safely pass None, it will be ignored.
                continue
            self.update_one(object_)

    def update_one(self, object_, **adapter_kw):
        """
        Push records from object in context with adapter kwargs

        `object_`: Instance of an object, usually a model object.

        `adapter_kw`: Pass extra kwargs to object.

        Usage:
        .. code-block:: python

            log_context.update(Request)
            log_context.update(Request, full_logs=True, display=True, ...)
        """
        deepupdate(self.data, adapt(object_, **adapter_kw))

    def remove(self, *keys):
        """
        Drop records from context.

        Inexistant records are ignored.
        """
        return self.remove_many(keys)

    def remove_many(self, keys):
        for k in keys:
            try:
                self.data.pop(k)
            except KeyError:
                pass

    def as_dict(self):
        """
        Return the context as a dict.
        """
        # Ensure thread local data is ready
        self.data

        items = {}
        # Squash all dict in stack and reverse order because
        # the first values in stack are the last inserted values and we need
        # to override values with the last inserted
        for d in reversed(self._stack):
            deepupdate(items, d)
        return items

    def items(self):
        return itertools.chain(*[self.as_dict().items()])

    def __call__(self, *objects, **fields):
        return UpdateContextManager(self, *objects, **fields)

    def cm_update_one(self, object_, **adapter_kw):
        return UpdateOneContextManager(self, object_, **adapter_kw)


class UpdateContextManager(object):
    def __init__(self, log_context, *objects, **fields):
        self.log_context = log_context
        self.objects = objects
        self.fields = fields

    def __enter__(self):
        # Ensure thread local data is ready
        self.log_context.data
        self.log_context._stack.insert(0, {})
        self.log_context.update(*self.objects, **self.fields)

    def __exit__(self, exc_type, exc_value, traceback):
        self.log_context._stack.pop(0)


class UpdateOneContextManager(object):
    def __init__(self, log_context, object_=None, **adapter_kw):
        self.log_context = log_context
        self.object_ = object_
        self.adapter_kw = adapter_kw

    def __enter__(self):
        # Ensure thread local data is ready
        self.log_context.data
        self.log_context._stack.insert(0, {})
        self.log_context.update_one(self.object_, **self.adapter_kw)

    def __exit__(self, exc_type, exc_value, traceback):
        self.log_context._stack.pop(0)


context = Context()


class AddContextFilter(logging.Filter):
    def __init__(self, name='', default=None):
        super(AddContextFilter, self).__init__(name)
        self.default = default or {}

    def filter(self, record):
        fields = dict(self.default, **context.as_dict())
        record.__dict__.update(fields)
        return True


class ExcInfoFilter(logging.Filter):
    def filter(self, record):
        record.exc_info = None
        return True


class AddContextFormatter(logging.Formatter):
    def format(self, record):
        msg = super(AddContextFormatter, self).format(record)
        context_str = ' '.join([
            '{}:{}'.format(k, v) for k, v in context.items()
        ])
        return '{msg} {context}'.format(msg=msg, context=context_str)


_adapter_mapping = {}


class AdapterNotFound(Exception):
    pass


def log_adapter(class_):
    def decorator(callable_):
        _adapter_mapping[class_] = callable_
        return callable_
    return decorator


def adapt(object_, **kwargs):
    for class_ in object_.__class__.__mro__:
        try:
            adapter = _adapter_mapping[class_]
            break
        except KeyError:
            continue
    else:
        raise AdapterNotFound("Can't log object of type %r" % type(object_))
    return adapter(object_, **kwargs)


class LazyAccessor(object):
    def __init__(self, instance, attrname):
        self.instance = instance
        self.attrname = attrname

    def get(self):
        return getattr(self.instance, self.attrname)

    def __unicode__(self):
        return unicode(self.get())

    def __str__(self):
        return str(self.get())

    def __repr__(self):
        return repr(self.get())



import os
import sys

from setuptools import setup
from setuptools.command.test import test as TestCommand

version = '1.12'

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['tests/']

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='pylogctx',
    version=version,
    description='Adding context to log records',
    long_description=README,
    classifiers=[
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords='logging context logging django',
    author="Lev Orekhov",
    author_email="lev.orekhov@gmail.com",
    url="https://github.com/novafloss/pylogctx",
    license="BSD",
    package_dir={'': 'src'},
    packages=['pylogctx'],
    cmdclass={'test': PyTest},
    include_package_data=True,
    zip_safe=False,
    tests_require=[
        'django>=1.11,<2' if sys.version_info[0] == 2 else 'django>=1.11',
        'mock>=2.0.0',
        'pytest>=2.9.0',
        'celery',
        'six'
    ],
)
